/********************************************************************
Octavia Petrovici, May 2001

  AngleRaysCirc.cpp
  Code for the disk mirror - rays intersection with an angular
  source of light.

  To run:
   % AngleRaysCirc [d][D] < file.input > output.ps
(d = debug flag)

Bounces a set of rays off a set of arbitrary circles.
The ray is specified by two points a and b, conventionally
starting at a and passing through b.  The circles are centered
on c[] radius r[].  The ray is first aimed at all circles, 
and the closest intersection is chosen. 
Then, the ray is aimed at all the other circles, 
closest intersection point chosen, etc, till it hits infinity 
(0 intersections)
If it hits a circle, a reflected ray is computed.  The point a 
is reflected to af, and the point where the ray intersects 
the circle is p. The closest p is chosen.
In the next iteration, the ray starts at p and passes through 
af, so (p,af) play the role that (a,b) did on the first iteration.

Input is the number of circles, their coordiantes and radii, the 
number of rays and the angular spread (start angle and the cone
of light, specified in degrees).

Output is a postscript file that contains a drawing of the 
intersections.


--------------------------------------------------------------------
This code is Copyright 2001 by Joseph O'Rourke.  It may be freely
redistributed in its entirety provided that this copyright notice is
not removed.
--------------------------------------------------------------------
*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>

#define EXIT_SUCCESS 0
#define EXIT_FAILURE 1
#define MARGIN 50
#define MINX MARGIN
#define MINY MARGIN
#define MAXX 612-MARGIN
#define MAXY 792-MARGIN
#define	Xi 0
#define	Yi 1
#define FUZZ 1
#define EPS  1e-12
#define NMAX 100
#define PI 3.14159 
#define FALSE 0
#define TRUE 1
#define DIM     2               /* Dimension of points */
typedef double  tPointd[DIM];   /* type double point */



/*----------Function Prototypes-------------*/
void	MoveTo( tPointd p );
void	LineTo( tPointd p );
void	SubVec( tPointd a, tPointd b, tPointd c );
void	AddVec( tPointd a, tPointd b, tPointd c );
bool	EqPointd( tPointd a, tPointd b );
double	Length2( tPointd v );
double	RadDeg( double x );
void 	PSHead( void );
void    ReScale(void);
void 	PSTail( void );
double	SQ( double x ) { return x * x; }
int 	IntRayCircle0( tPointd a, tPointd b, double r, tPointd p );
int     IntRayCircle( tPointd a, tPointd b, tPointd c, double r, 
					 tPointd p, tPointd af );
void	PointAssign( tPointd a, tPointd b );
bool 	Iterate( tPointd a, tPointd b );
void 	DrawCircles( void );
void	PrintInfo(void);
double  SqDist(tPointd a, tPointd b);

/*------------------------------------------*/
/* Global variables. */
tPointd c[10000];				//centers of circles
double r[10000];				//radii of circles
double rLimit[4];
tPointd cLimit[4];

//rescaling variables:
double transX, transY;
double scaleX, scaleY;

double startangle;
double angle;
int NC;							//number of circles
int NR;							//number of rays
int nbmax = 10;					//number of bounces

bool debug = FALSE;
bool mydebug = FALSE;
#define IFD if (debug)
#define IFM if (mydebug)

int main ( int argc, char *argv[] ){
	tPointd a, b;
	tPointd sCoord;
	tPointd aa[10000];
	tPointd bb[10000];
	bool above = FALSE;
	double delta = 0.01;
	bool forward = TRUE;
	
	if (argc == 2 && argv[1][0] ==  'd'){
		debug = TRUE;
		mydebug = TRUE;
	}       
	if (argc == 2 && argv[1][0] ==  'D'){
		mydebug = TRUE;
	}	
	PSHead();	
	printf("%% Enter the number of circles you wish to create:\n");
	scanf("%d", &NC);
	if (NC>10000){
		printf("10 000 is the maximum allowed.");
		exit(0);
	}
	printf("%%Enter each circle's xy coordinate and radius:\n");
	for (int j = 0; j<NC; j++){
		scanf("%lf %lf %lf", &c[j][0], &c[j][1], &r[j]);
	}      

	//rescaling part
	double minX, minY, maxX, maxY;      
	minX = c[0][0]-r[0];
	maxX = c[0][0]+r[0];
	minY = c[0][1]-r[0];
	maxY = c[0][1]+r[0];
	
	for (j = 1; j<NC; j++){
		if (minX > c[j][0]-r[j])
			minX = c[j][0]-r[j];
		if (maxX < c[j][0]+r[j])
			maxX = c[j][0]+r[j];
		if (minY > c[j][1]-r[j])
			minY = c[j][1]-r[j];
		if (maxY < c[j][1]+r[j])
			maxY = c[j][1]+r[j];
	}
	if (minX > 0)
		minX = 0;
	if (minY > 0)
		minY = 0;

	IFM printf("%% MinX is %lf, MaxX is %lf, \n %%MinY is %lf, MaxY is %lf\n", minX, maxX, minY, maxY);

	double spanX = maxX-minX>0?maxX-minX:minX-maxX;
	double spanY = maxY-minY>0?maxY-minY:minY-maxY;
	double trans = (MAXX-MINX)/spanX<(MAXY-MINY)/spanY?(MAXX-MINX)/spanX:(MAXY-MINY)/spanY;
	IFM printf("%% spanX is %lf and spanY is %lf\n", spanX, spanY);
	IFM printf("%% scaling factor is %lf \n", trans);
	transX = (0-minX)*trans+MARGIN;
	transY = (0-minY)*((MAXY-MINY)/spanY)+MARGIN;
	scaleX = trans;
	scaleY = trans;	
	cLimit[0][0] = minX + spanX/2;
	cLimit[0][1] = maxY + 11*spanX;
	rLimit[0] = 10*spanX;	
	cLimit[1][0] = minX + spanX/2;
	cLimit[1][1] = minY - 11*spanX;
	rLimit[1] = 10*spanX;	
	cLimit[2][0] = minX - 11*spanY;
	cLimit[2][1] = minY + spanY/2;
	rLimit[2] = 10*spanY; 
	cLimit[3][0] = maxX + 11*spanY;
	cLimit[3][1] = minY + spanY/2;
	rLimit[3] = 10*spanY;	
	ReScale();	
	//end rescaling part

	printf("%%NC is %d \n", NC);	
	DrawCircles();
	printf("%%Enter xy coordinates for light source:\n");
	scanf("%lf %lf", &sCoord[0], &sCoord[1]);
	for ( j=0; j<NC; j++){
		if (sqrt((sCoord[0]-c[j][0])*(sCoord[0]-c[j][0])+
		(sCoord[1]-c[j][1])*(sCoord[1]-c[j][1]))<r[j]){
			printf("%%Source in the circle!!! Please choose a point source external to all circles.");
         exit(0);
       }
	}     
	printf("%%Enter number of rays:\n");
	scanf("%d", &NR);	
	printf("%%Enter positive start angle in degrees:\n");
	scanf("%lf", &startangle);
	printf("%%Enter positive angle in degrees:\n");	
	scanf("%lf", &angle);	

	printf("%% Maximum number of ray bounces (nbmax):\n");
	scanf("%d", &nbmax);

	double aaux;	
	for (j = 0; j<NR; j++){
		if(NR==1)
			aaux = startangle;
		else
			aaux = angle/(NR-1)*j+startangle;		
		PointAssign(aa[j], sCoord);		
		bb[j][0] = cos(aaux*PI/180.) + aa[j][0];
		bb[j][1] = sin(aaux*PI/180.) + aa[j][1];
	}		
	printf("%% aaux (deg) = %g\n", aaux);
        printf("newpath ");
        printf("%g %g  7 0 360 arc", sCoord[0], sCoord[1]); 
        printf(" closepath stroke\n");
	for ( j = 0; j<NR; j++){
		PointAssign(a, aa[j]);
		PointAssign(b, bb[j]);
		Iterate( a, b );
	}	
	PrintInfo();	
	PSTail();
	return EXIT_SUCCESS;
}

bool Iterate( tPointd a, tPointd b ){
	tPointd p, af;
	tPointd p0, af0;
	tPointd p1, af1;
	int nsoln = -1;
	int i;
	bool above = FALSE;       
	double distaux = -1;
	double currdist = 1000000;
	int starti = -1;
	int newstarti = -1;
	int nbounce = 0;
	printf("newpath\n");
	MoveTo( a );
	PointAssign( p1, a );
	PointAssign( af1, b );
	i = 0;
	bool intersect = true;

	while (intersect && (nbounce < nbmax)){
		intersect = false;
		currdist = 1000000;
		nbounce++;
		for (i = 0; i<NC; i++){
			if (starti != i){
				IFM printf("%%Intersect circle %d (%lf %lf %lf)\n %%with ray ((%lf %lf) (%lf %lf))\n", i, c[0][i], c[1][i], r[i], p1[0], p1[1], af1[0], af1[1]);
				nsoln = IntRayCircle( p1, af1, c[i], r[i], p0, af0 );
				IFM printf("%%nsoln is %d \n", nsoln);
				IFM printf("%% New intersection point is (%lf %lf)\n", p0[0], p0[1]);
					if ( nsoln != 0 ){
						distaux = SqDist(p1, p0);
							if((distaux<currdist)&&(starti!=i)&&(distaux>0.0)){
								intersect = true;
								newstarti = i;
								PointAssign( p, p0 );
								PointAssign( af, af0 );						              
								currdist = distaux;	       
							}
							else
								IFM{
									if (starti == i)
										printf("%%Same i: %d = %d \n", starti, i);
									else
										printf("%%Distance af1, p0 is bigger then currdist, %lf > %lf\n", distaux, currdist);
							}
					}
			}
		}		
		if (intersect){
			starti = newstarti;
			PointAssign( p1, p );
			PointAssign( af1, af );
			LineTo( p1 );	 
		}
	}
	if ( !intersect ) {
	  LineTo( af1 );
 for (int ll = 0; ll<4; ll++){
   nsoln = IntRayCircle( p1, af1, cLimit[ll], rLimit[ll], p0, af0 );
 if (nsoln != 0)
 break;
 }
	LineTo(p0);
}
	printf("Rcolor\n");
	printf("stroke\n");
	return above;
}

/*--------------------------------------------------------------

 Input:
 a = ray origin
 b = ray through point
 c = center of circle
 r = radius of circle
 Output:
 returns the number of intersections
 p = first point of intersection
 af = reflection of point a about the line through c and p.
 
  Shift so c is origin and call IntRayCircle0.
  Returns reflected(flipped) a-point, af.  So new ray is p->af.

--------------------------------------------------------------*/

int IntRayCircle( tPointd a, tPointd b, tPointd c, double r, 
				 tPointd p, tPointd af )
{
	tPointd a0, b0, p0, af0;
	int nsoln = -1;
	double costheta, sintheta;
	tPointd ar, pr;  /* Rotated points */
	tPointd arf;     /* Rotated/flipped (reflected) */
	
	IFD printf("IntRayCircle: a=(%g,%g),b=(%g,%g),c=(%g,%g),r=%g\n",
		a[0],a[1],b[0],b[1],c[0],c[1],r);
	if ( r == 0.0 ) {
		printf("Error in IntRayCircle: r = 0\n");
		exit( EXIT_FAILURE );
	}
	
	SubVec( a, c, a0 );
	SubVec( b, c, b0 );
	nsoln = IntRayCircle0( a0, b0, r, p0 );
	if ( nsoln == 0 ) {
		IFD printf("IntRayCircle: Missed the circle!\n");
		return nsoln;
	}
	/* Otherwise go ahead and compute the reflected ray. */
	
	/* Compute theta around circle of point p0. */
	costheta = p0[0] / r;
	sintheta = p0[1] / r;
	IFD printf("IntRayCircle: costheta=%g, sintheta=%g\n", costheta, sintheta);
	IFD printf("IntRayCircle: theta=%g\n", RadDeg( atan2(sintheta,costheta) ) );
	
	/* Rotate a0 and p0 by theta to place p0 on x-axis. */
	pr[0] =  costheta * p0[0] + sintheta * p0[1];
	pr[1] = -sintheta * p0[0] + costheta * p0[1];
	if ( fabs( pr[0] - r ) > FUZZ ) {
		printf("Error in IntRayCircle: pr[0]=%lf should be r=%lf\n", pr[0], r);
		exit( EXIT_FAILURE );
	}
	ar[0] =  costheta * a0[0] + sintheta * a0[1];
	ar[1] = -sintheta * a0[0] + costheta * a0[1];
	IFD printf("IntRayCircle: pr = (%g,%g)\n", pr[0], pr[1] );
	IFD printf("IntRayCircle: ar = (%g,%g)\n", ar[0], ar[1] );
	
	/* Flip (reflect) point a in x-axis, to get reflected ray. */
	arf[0] =  ar[0];
	arf[1] = -ar[1];
	/* Rotate back. */
	af[0] =  costheta * arf[0] - sintheta * arf[1];
	af[1] =  sintheta * arf[0] + costheta * arf[1];
	IFD printf("IntRayCircle: arf = (%g,%g)\n", arf[0], arf[1] );
	IFD printf("IntRayCircle: af  = (%g,%g)\n", af[0], af[1] );
	/* This point af is really in the origin-centered coord. system. */
	af0[0] = af[0];
	af0[1] = af[1];
	
	/* Put points back into given coordinate system. */
	AddVec( p0,  c, p );
	AddVec( af0, c, af);
	IFD printf("IntRayCircle: p = (%g,%g), af=(%g,%g)\n", p[0], p[1], af[0], af[1] );
	return nsoln;
}

/*--------------------------------------------------------------
Assume that the circle is centered on the origin.
Returns the number of solutions, and the first point of intersection.
--------------------------------------------------------------*/

int IntRayCircle0( tPointd a, tPointd b, double r, tPointd p ){
	double mb, disc, denom;
	double r2;
	double t1, t2, t;      /* t is the one of {t1,t2} for the 1st inter. */
	int nsoln = -1;
	int i;
	
	IFD printf("IntRayCircle0: a=(%g,%g),b=(%g,%g),r=%g\n",
		a[0],a[1],b[0],b[1],r);
	r2 = SQ(r);
	mb = SQ(a[0]) + SQ(a[1]) - a[0]*b[0] - a[1]*b[1];
			
	disc = -SQ(a[1]*b[0]) + 2*a[0]*a[1]*b[0]*b[1] - SQ(a[0]*b[1]) + SQ(a[0])*r2 + 
		SQ(a[1])*r2 - 2*a[0]*b[0]*r2 + SQ(b[0])*r2 - 2*a[1]*b[1]*r2 + SQ(b[1])*r2;
    
	denom = SQ(a[0]) + SQ(a[1]) - 2*a[0]*b[0] + SQ(b[0]) - 2*a[1]*b[1] + SQ(b[1]);
	
	IFD printf("r2=%g, mb=%g, disc=%g, denom=%g\n", r2, mb, disc, denom);	
	if ( denom == 0.0 ) {
		printf("Error in IntRayCircle0: denom == 0\n");
		nsoln = 0;
		return nsoln;
	}
	if ( disc == 0.0 ) {
		IFD printf("IntRayCircle0: One solution because disc == 0\n");
		t = mb / denom;
		nsoln = 1;
	}
	if ( disc < 0.0 ) {
		IFD printf("IntRayCircle0: No solutions because negative disc\n");
		nsoln = 0;
		return nsoln;
	}
	
	t1 = ( mb - sqrt( disc ) ) / denom;
	t2 = ( mb + sqrt( disc ) ) / denom;
	IFD printf("IntRayCircle0: t1=%g, t2=%g\n", t1, t2 );
	
	if ( t1 > t2 ) 
		printf("Error in IntRayCircle0: t1 > t2 should never happen!\n");
	
	if ( t1 < 0.0  &&  t2 < 0.0 ) {
		IFD printf("IntRayCircle0: Both intersection on negative ray\n");
		nsoln = 0;
		return nsoln;
	}
	
	nsoln = 2;
	if ( t1 < 0.0 )
		t = t2;
	else /* 0.0 < t1 < t2 */
		t = t1;
	for( i = 0; i < DIM; i++ ) 
		p[i] = a[i] + t * ( b[i] - a[i] );
	IFD printf("IntRayCircle0: Using t = %g: p = (%g,%g)\n", t, p[0], p[1] );
	return nsoln;
}

void   MoveTo( tPointd p ){
	printf("%10.5lf %10.5lf moveto\n", p[Xi], p[Yi] );
}

void   LineTo( tPointd p ){
	printf("%10.5lf %10.5lf lineto\n", p[Xi], p[Yi] );
}

double   Length2( tPointd v ){
	int   i;
	double   ss;
	ss = 0;
	for ( i=0; i < DIM; i++ )
		ss += v[i] * v[i];
	return ss;
}

double   RadDeg( double x ){
	return 180 * x / PI;
}

/* a - b -> c */
void   SubVec( tPointd a, tPointd b, tPointd c ){
	int   i;
	for ( i=0; i < DIM; i++ ) {
		c[i] = a[i] - b[i];
	}
}

/* a + b -> c */
void   AddVec( tPointd a, tPointd b, tPointd c ){
	int   i;
	for ( i=0; i < DIM; i++ ) {
		c[i] = a[i] + b[i];
	}
}

void PSHead( void ){
	printf("%%!PS\n");
	printf(".002 .002 setlinewidth\n");
	printf("%% ------------------------ Color ---------------------%%\n");
	printf("/Rcolor { 1.0 0 0.45 setrgbcolor }def\n");	
}

void ReScale(void){
	printf("%lf %lf translate\n", transX, transY ); 
	printf("%lf %lf scale\n", scaleX, scaleY ); /* 1 -> 2 inches */
	printf("newpath\n");
}


void PSTail( void )
{
	printf("\nshowpage\n\n");
	printf("%%EOF\n");
}

/* a = b */
void PointAssign( tPointd a, tPointd b )
{
	int i;
	for ( i=0; i < DIM; i++ ) {
		a[i] = b[i];
	}
}

void DrawCircles( void ) {
	for (int j = 0; j<NC; j++)
		printf("newpath\t%lf %lf %lf 0 360 arc\tclosepath stroke\n", c[j][0], c[j][1], r[j]); 
}

/* Use PrintInfo to print any useful information in the .ps file */
void PrintInfo(void){
	printf("/Times-Roman findfont   %% Get the basic font\n");
	printf("%lf scalefont            %% Scale the font to 20 points\n", 11/scaleX);
	printf("setfont                 %% Make it the current font\n newpath                 %% Start a new path\n");
	printf("%lf %lf moveto\n", -transX/scaleX+MARGIN/(2*scaleX), -transY/scaleY+MARGIN/(2*scaleY));
	printf("(There are %d circles, %d rays; start at angle %lf deg, for %lf deg. X dist circle 1->2: %lf) show\n",NC, NR, startangle, angle, c[1][0]-c[0][0]-r[0]-r[1]);       
}

double SqDist(tPointd a, tPointd b)
{
	return((a[0]-b[0])*(a[0]-b[0])+(a[1]-b[1])*(a[1]-b[1]));
}



