/********************************************************************
Octavia Petrovici, May 2001

  AngleRaysSeg.cpp
  Code for the segment mirror - rays intersection with an angular 
  source of light.

  To run:
  % AngleRaysSeg [d][D] < file.input > output.ps
  (d = debug flag)

  Bounces a set of rays off a set of arbitrary segments.

The rays are specified by two endpoints each, a and b, c and d, 
onventionally starting at a(/c) and passing through b(/d).  
The coordinates are stored in sStart[] and sEnd[].
The ray is first aimed at all segments, and the closest 
intersection is chosen. 
Then, the ray is aimed at all the other segments, 
closest intersection point chosen, etc, till it hits infinity 
(0 intersections)
If it hits a segment, a reflected ray is computed.  The point a 
is reflected to af, and the point where the ray intersects 
the circle is p. The closest p is chosen.
In the next iteration, the ray starts at p and passes through 
af, so (p,af) play the role that (a,b) did on the first iteration.

Input is the number of segments, their coordiantes and radii, the 
number of rays and the angular spread (start angle and the cone
of light, specified in degrees).

Output is a postscript file that contains a drawing of the 
intersections.
--------------------------------------------------------------------
This code is Copyright 2001 by Joseph O'Rourke.  It may be freely
redistributed in its entirety provided that this copyright notice is
not removed.
--------------------------------------------------------------------
*/

#include <stdio.h>
#include <math.h>
//#include <util.h>
#include <stdlib.h>
#include <time.h>
#define EXIT_SUCCESS 0
#define EXIT_FAILURE 1
#define MARGIN 50
#define MINX MARGIN
#define MINY MARGIN
#define MAXX 612-MARGIN
#define MAXY 792-MARGIN
#define	Xi 0
#define	Yi 1
#define FUZZ 1e-2
#define EPS  1e-12
#define NMAX 100
#define bool int
#define TRUE 1
#define FALSE 0
#define DIM     2               /* Dimension of points */

int nbmax = 10;
typedef double  tPointd[DIM];   /* type double point */

/*----------Function Prototypes  -------------*/
void	MoveTo( tPointd p );
void	LineTo( tPointd p );
void	SubVec( tPointd a, tPointd b, tPointd c );
void	AddVec( tPointd a, tPointd b, tPointd c );
bool	EqPointd( tPointd a, tPointd b );
double	Length2( tPointd v );
double	RadDeg( double x );
void 	PSHead( void );
void    ReScale(void);
void 	PSTail( void );
double	SQ( double x ) { return x * x; }
int 	IntRayCircle0( tPointd a, tPointd b, double r, tPointd p );
int     IntRayCircle( tPointd a, tPointd b, tPointd c, double r, tPointd p, tPointd af );
void	PointAssign( tPointd a, tPointd b );
bool 	Iterate( tPointd a, tPointd b );
void 	DrawSegments( void );
void	PrintInfo(void);
double  SqDist(tPointd a, tPointd b);
double  Dist(tPointd a, tPointd b);
int 	RaySegInt( tPointd a, tPointd b, tPointd c, tPointd d, tPointd p, tPointd af );
char	ParallelInt( tPointd a, tPointd b, tPointd c, tPointd d, tPointd p );
bool	Between( tPointd a, tPointd b, tPointd c );
bool	OnRay( tPointd a, tPointd b, tPointd c );
void	Assigndi( tPointd p, tPointd a );
int 	Collinear( tPointd a, tPointd b, tPointd c );
int     AreaSign( tPointd a, tPointd b, tPointd c );
int     ReflectRay(tPointd a, tPointd p, tPointd c, tPointd d, tPointd af);

/*-------------------------------------------------------------------*/

/* Global variables. */
tPointd sStart[10000];
tPointd sEnd[10000];
double r[100];
double rLimit[4];
tPointd cLimit[4];

//rescaling variables
double transX, transY;
double scaleX, scaleY;

//start point and end point for the range 
//of rays 
tPointd startpoint;
tPointd endpoint;

double startangle, angle;

int NS;				//number of segments
int NR;				//number of rays

bool debug = FALSE;
bool mydebug = FALSE;

#define IFD if (debug)
#define IFM if (mydebug)

int main ( int argc, char *argv[] ){
  tPointd a, b;
  tPointd aa[30000];
  tPointd bb[30000];	
  bool above = FALSE;
  double delta = 0.01;
  bool forward = TRUE;
	
  if (argc == 2 && argv[1][0] ==  'd'){
      debug = TRUE;
      mydebug = TRUE;

    }
  if (argc == 2 && argv[1][0] ==  'D')
    {
      mydebug = TRUE;
    }	
  PSHead();	
  printf("%% Number of segments and their coords:\n");
  scanf("%d", &NS);
  for (int j = 0; j<NS; j++)
  scanf("%lf %lf %lf %lf", &sStart[j][0], &sStart[j][1], &sEnd[j][0], &sEnd[j][1]);

  //rescaling part
  double minX, minY, maxX, maxY;
  minX = sStart[0][0];
  maxX = sStart[0][0];
  minY = sStart[0][1];
  maxY = sStart[0][1];

  for (int j = 1; j<NS; j++){
    if (minX > sStart[j][0])
      minX = sStart[j][0];
    if (minX > sEnd[j][0])
      minX = sEnd[j][0];
    if (maxX < sStart[j][0])
      maxX = sStart[j][0];
    if (maxX < sEnd[j][0])
      maxX = sEnd[j][0];
    if (minY > sStart[j][1])
      minY = sStart[j][1];
    if (minY > sEnd[j][1])
      minY = sEnd[j][1];
    if (maxY < sStart[j][1])
      maxY = sStart[j][1];
    if (maxY < sEnd[j][1])
      maxY = sEnd[j][1];
  }

  if (minX > 0) minX = 0;
  if (minY > 0) minY = 0;

  IFM printf("%% MinX is %lf, MaxX is %lf, \n %%MinY is %lf, MaxY is %lf\n", minX, maxX, minY, maxY);
  double spanX = maxX-minX>0?maxX-minX:minX-maxX;
  double spanY = maxY-minY>0?maxY-minY:minY-maxY;
  double trans = (MAXX-MINX)/spanX<(MAXY-MINY)/spanY?(MAXX-MINX)/spanX:(MAXY-MINY)/spanY;
	
  IFM printf("%% spanX is %lf and spanY is %lf\n", spanX, spanY);
  IFM printf("%% scaling factor is %lf \n", trans);

  transX = (0-minX)*trans+MARGIN;
  transY = (0-minY)*((MAXY-MINY)/spanY)+MARGIN;
  scaleX = trans;
  scaleY = trans;
  cLimit[0][0] = minX + spanX/2;
  cLimit[0][1] = maxY + 11*spanX;
  rLimit[0] = 10*spanX;
  cLimit[1][0] = minX + spanX/2;
  cLimit[1][1] = minY - 11*spanX;
  rLimit[1] = 10*spanX;
  cLimit[2][0] = minX - 11*spanY;
  cLimit[2][1] = minY + spanY/2;
  rLimit[2] = 10*spanY;
  cLimit[3][0] = maxX + 11*spanY;
  cLimit[3][1] = minY + spanY/2;
  rLimit[3] = 10*spanY;
  ReScale();

  //end rescaling part
  DrawSegments();
  printf("%%Enter source coordinates:\n");
  scanf("%lf %lf", &startpoint[0], &startpoint[1]);	
  printf("%%Enter number of rays:\n");
  scanf("%d", &NR);
  printf("%%Enter positive start angle in degrees:\n");
  scanf("%lf", &startangle);
  printf("%%Enter positive angle in degrees:\n");
  scanf("%lf", &angle);	
  printf("%% Maximum number of ray bounces (nbmax):\n");
  scanf("%d", &nbmax);
     
  double aaux; //Auxilary angle.
	
  for (int j = 0; j<NR; j++)
   {
     if (NR == 1) 
         aaux = startangle;
     else
         aaux = angle/(NR-1)*j+startangle;
     PointAssign(aa[j], startpoint);
     bb[j][0] = cos(aaux*M_PI/180.) + aa[j][0];
     bb[j][1] = sin(aaux*M_PI/180.) + aa[j][1];
   }

  for (int j = 0; j<NR; j++)
   {
     PointAssign(a, aa[j]);
     PointAssign(b, bb[j]);
     Iterate( a, b );
    }
     PrintInfo();
     PSTail();
     return EXIT_SUCCESS;
}


bool Iterate( tPointd a, tPointd b )
{
	tPointd p, af;
	tPointd p0, af0;
	tPointd p1, af1;
	int nsoln = -1;
	int i;
	bool above = FALSE;	
	double distaux = -1;
	double currdist = 1000000;
	int starti = -1;
	int newstarti = -1;
	int nbounce = 0;
	
        printf("newpath\n");
	MoveTo( a );
	PointAssign( p1, a );
	PointAssign( af1, b );
	i = 0;	
	bool intersect = true;

	while (intersect && (nbounce < nbmax)){
	   intersect = false;
	   currdist = 1000000;

	   nbounce++;
	   for (int i = 0; i<NS; i++)
	     {
	       if (starti != i)
		 {
		   IFM printf("%%Intersect segment %d (%lf %lf %lf %lf)\n %%with ray ((%lf %lf) (%lf %lf))\n", i, sStart[0][i], sStart[1][i], sEnd[0][i], sEnd[1][i], p1[0], p1[1], af1[0], af1[1]);
		   nsoln = RaySegInt( p1, af1, sStart[i], sEnd[i], p0, af0 );
		   IFM printf("%%nsoln is %d \n", nsoln);
		   IFM printf("%% New intersection point is (%lf %lf)\n", p0[0], p0[1]);
		   if ( nsoln != 0 )
		     {
		       distaux = SqDist(p1, p0);
		       if((distaux<currdist)&&(starti!=i)&&(distaux>0.0))
			 {
			   intersect = true;
			   newstarti = i;
			   PointAssign( p, p0 );
			   PointAssign( af, af0 );
			   currdist = distaux;				       				 }

		       else
			 IFM{
			     if (starti == i)
			       printf("%%Same i: %d = %d \n", starti, i);
			     else
			       printf("%%Distance af1, p0 is bigger then currdist, %lf > %lf\n", distaux, currdist);
			   }
		     }
		 }
	     }//end NS-for		
	   if (intersect){
			starti = newstarti;
			PointAssign( p1, p );
			PointAssign( af1, af );
			LineTo( p1 );	  			
		}
	}// end-while



	if ( !intersect ) {
	LineTo( af1 );
	for (int ll = 0; ll<4; ll++)
	  {
	    nsoln = IntRayCircle( p1, af1, cLimit[ll], rLimit[ll], p0, af0 );
	    if (nsoln != 0)
	      break;
	  }
	LineTo(p0);
	}
	printf("Rcolor\n");
	printf("stroke\n");
	return above;
}

/*---------------------------------------------------------------------
Reflect Ray: Reflects the ray a->p (where p is on the segment cd) off
segment cd. First translates p tp p0, the origin, then rotates so that 
a is a0 and cd is orizontal. Computes the reflected af0, then rotates 
and translates back. The reflected a is in af.
---------------------------------------------------------------------*/

int ReflectRay(tPointd a, tPointd p, tPointd c, tPointd d, tPointd af)
{
	tPointd a0, p0, c0, d0, af0;
	double costheta, sintheta;
	tPointd ar, cr, dr;  /* Rotated points */
	tPointd arf;     /* Rotated/flipped (reflected) */		
	SubVec( a, p, a0 );
	SubVec( p, p, p0 );
	SubVec( c, p, c0 );
	SubVec( d, p, d0 );
	       
	/* Compute theta around circle of point p0. */
	costheta = (d[0]-c[0]) / Dist(c, d);
	sintheta = (d[1]-c[1]) / Dist(c, d);
	IFD printf("Reflect Ray: costheta=%g, sintheta=%g\n", costheta, sintheta);
	IFD printf("Reflect Ray: theta=%g\n", RadDeg( atan2(sintheta,costheta) ) );

       	/* Rotate a0 and p0 by theta to place p0 on x-axis. */
	cr[0] =  costheta * c0[0] + sintheta * c0[1];
	cr[1] = -sintheta * c0[0] + costheta * c0[1];
	dr[0] =  costheta * d0[0] + sintheta * d0[1];
	dr[1] = -sintheta * d0[0] + costheta * d0[1];	
	ar[0] =  costheta * a0[0] + sintheta * a0[1];
	ar[1] = -sintheta * a0[0] + costheta * a0[1];	
	/* Flip (reflect) point a in y-axis, to get reflected ray. */
	arf[0] = - ar[0];
	arf[1] = ar[1];	
	/* Rotate back. */
	af[0] =  costheta * arf[0] - sintheta * arf[1];
	af[1] =  sintheta * arf[0] + costheta * arf[1];
	IFD printf("RayReflect: arf = (%g,%g)\n", arf[0], arf[1] );
	IFD printf("RayReflect: af  = (%g,%g)\n", af[0], af[1] );
	/* This point af is really in the origin-centered coord. system. */
	af0[0] = af[0];
	af0[1] = af[1];
	/* Put points back into given coordinate system. */	
	AddVec( af0, p, af);
	//IFD printf("IntRayCircle: p = (%g,%g), af=(%g,%g)\n", p[0], p[1], af[0], af[1] );
	return 0;
}

/*---------------------------------------------------------------------
RaySegInt: Finds the point of intersection p between a ray starting
at a and passing through b, and the closed segment cd.
Returns p and a char with the following meaning:
'e': The ray collinearly overlaps the segment, sharing a point (or more).
'v': The ray intersects the segment at an endpoint,
but 'e' doesn't hold.
'r': The root of the ray is on the segment, but 'e' doesn't hold.
'1': The ray properly intersects the segment (i.e., they share a point and
neither 'v' nor 'e' holds).
'0': The ray misses the segment entirely (i.e., they share no points).
---------------------------------------------------------------------*/

int	RaySegInt( tPointd a, tPointd b, tPointd c, tPointd d, tPointd p, tPointd af )

{
	double  s, t;       /* The two parameters of the parametric eqns. */
	double num, denom;  /* Numerator and denoninator of equations. */
	char code = '?';    /* Return char characterizing intersection. */     
	denom = a[Xi] * (double)( d[Yi] - c[Yi] ) +
		b[Xi] * (double)( c[Yi] - d[Yi] ) +
		d[Xi] * (double)( b[Yi] - a[Yi] ) +
		c[Xi] * (double)( a[Yi] - b[Yi] );

	/* If denom is zero, then segments are parallel: handle separately. */
	IFD printf("%%denom=%lf\n", denom);
	if (denom == 0.0)
	{
	  code = ParallelInt(a, b, c, d, p);
	  if (code == 'e')
	    {
	      PointAssign(p, b);
	      PointAssign(af, a);
	      return 1;
	    }
	  else 
	    return 0;
	}
	num =    a[Xi] * (double)( d[Yi] - c[Yi] ) +
		c[Xi] * (double)( a[Yi] - d[Yi] ) +
		d[Xi] * (double)( c[Yi] - a[Yi] );
	IFD printf("%%num=%lf, denom=%lf\n", num, denom);
	if ( (num == 0.0) || (num == denom) ) {
	  if ( OnRay(a,b,c) || OnRay(a,b,d) )
	    {
	      code = 'v';
	      PointAssign(p, b);
	      ReflectRay(a, p, c, d, af);
	      //reflect the ray:			
	    }
	  else if ( Between(c,d,a) )
	    {
	      code = 'r';
	      printf("%%Error in RaySeg: origin on the segment!\n");
	      return 0;		
		}
	  else printf("%%Error in RaySeg: Undefined case!\n");		
	}

	s = num / denom;
	IFD printf("%%num=%lf, denom=%lf, s=%lf\n", num, denom, s);
	num = -( a[Xi] * (double)( c[Yi] - b[Yi] ) +
		b[Xi] * (double)( a[Yi] - c[Yi] ) +
		c[Xi] * (double)( b[Yi] - a[Yi] ) );

	IFD printf("%%num=%lf, denom=%lf\n", num, denom);
	if ( (num == 0.0) || (num == denom) ) {
	  if ( OnRay(a,b,c) || OnRay(a,b,d) )
	    {
	      code = 'v';
	      ReflectRay(a, b, c, d, af);
	    }
	  else if ( Between(c,d,a) )
		{
		  code = 'r';
		  printf("%%Error in RaySeg: origin on the segment!\n");
		  return 0;
		}
	  else printf("%%Error in RaySeg: Undefined case!\n");
	}
	t = num / denom;
	IFD printf("%%num=%lf, denom=%lf, t=%lf\n", num, denom, t);
	
	/* The parameter s runs along the ray ab.
	The parameter t runs along the segment cd. */
	if      ( (0.0 < s) &&
		(0.0 < t) && (t < 1.0) )
		code = '1';
	else if ( (0.0 > s) ||
		(0.0 > t) || (t > 1.0) )
	  {
		code = '0';
		return 0;
	  }  	
	p[Xi] = a[Xi] + s * ( b[Xi] - a[Xi] );
	p[Yi] = a[Yi] + s * ( b[Yi] - a[Yi] );  	
	ReflectRay(a, p, c, d, af);  	
 	//last check 

	if (code == 'e' || code == '1')
 		return 1;
	else
 		return 0;
}


char	ParallelInt( tPointd a, tPointd b, tPointd c, tPointd d, tPointd p )
{

	//printf("%%ParallelInt called\n");
	/* Only return e if both c and d lie along line containing ab. */
  if ( !(Collinear( a, b, c) && Collinear( a, b, d)) )
    return '0';
        
  if ( OnRay( a, b, c ) ) {
    Assigndi( p, c );
    return 'e';
  }
  if ( OnRay( a, b, d ) ) {
    Assigndi( p, d );
    return 'e';
  }
  if ( Between( c, d, a ) ) {
    Assigndi( p, a );
    return 'e';
  }
  return '0';
}

void	Assigndi( tPointd p, tPointd a )
{
  int i;
  for ( i = 0; i < DIM; i++ )
    p[i] = a[i];
}

/*---------------------------------------------------------------------
Returns TRUE iff point c lies on the closed segement ab.
Checks for collinearity.
---------------------------------------------------------------------*/

bool    Between( tPointd a, tPointd b, tPointd c )
{  	
  if (!Collinear(a,b,c)) return false;
  /* If ab not vertical, check betweenness on x; else on y. */
  if ( a[Xi] != b[Xi] )
    return ((a[Xi] <= c[Xi]) && (c[Xi] <= b[Xi])) ||
      ((a[Xi] >= c[Xi]) && (c[Xi] >= b[Xi]));
  else
    return ((a[Yi] <= c[Yi]) && (c[Yi] <= b[Yi])) ||
      ((a[Yi] >= c[Yi]) && (c[Yi] >= b[Yi]));
}

/*---------------------------------------------------------------------
Returns TRUE iff point c lies on the closed ray ab.
Checks for collinearity.
---------------------------------------------------------------------*/

bool    OnRay( tPointd a, tPointd b, tPointd c )
{  	
  if (!Collinear(a,b,c)) return false;
  /* If ab not vertical, check to correct side of root a. */
  if ( a[Xi] != b[Xi] )
     return ((a[Xi] < b[Xi]) && (a[Xi] <= c[Xi])) ||
      ((a[Xi] > b[Xi]) && (a[Xi] >= c[Xi]));
  else
    return ((a[Yi] < b[Yi]) && (a[Yi] <= c[Yi])) ||
      ((a[Yi] > b[Yi]) && (a[Yi] >= c[Yi]));
}

int Collinear( tPointd a, tPointd b, tPointd c )
{
  return AreaSign( a, b, c ) == 0;
}

int     AreaSign( tPointd a, tPointd b, tPointd c )
{
  double area2;         
  area2 = ( b[0] - a[0] ) * (double)( c[1] - a[1] ) -
    ( c[0] - a[0] ) * (double)( b[1] - a[1] );

  /* The area should be an integer. */
  if      ( area2 >  0.5 ) return  1;
  else if ( area2 < -0.5 ) return -1;
  else                     return  0;
}

/*--------------------------------------------------------------
 Input:
 a = ray origin
 b = ray through point
 c = center of circle
 r = radius of circle
 Output:
 returns the number of intersections
 p = first point of intersection
 af = reflection of point a about the line through c and p.
 
 Shift so c is origin and call IntRayCircle0.
 Returns reflected(flipped) a-point, af.  So new ray is p->af.
--------------------------------------------------------------*/
int IntRayCircle( tPointd a, tPointd b, tPointd c, double r,                                   tPointd p, tPointd af )
{
	tPointd a0, b0, p0, af0;
	int nsoln = -1;
	double costheta, sintheta;
	tPointd ar, pr;  /* Rotated points */
	tPointd arf;     /* Rotated/flipped (reflected) */
	
	IFD printf("IntRayCircle: a=(%g,%g),b=(%g,%g),c=(%g,%g),r=%g\n",
		a[0],a[1],b[0],b[1],c[0],c[1],r);
	if ( r == 0.0 ) {
		printf("Error in IntRayCircle: r = 0\n");
		exit( EXIT_FAILURE );
	} 	
	SubVec( a, c, a0 );
	SubVec( b, c, b0 );
	nsoln = IntRayCircle0( a0, b0, r, p0 );
	if ( nsoln == 0 ) {
		IFD printf("IntRayCircle: Missed the circle!\n");
		return nsoln;
	}
	/* Otherwise go ahead and compute the reflected ray. */
	
	/* Compute theta around circle of point p0. */
	costheta = p0[0] / r;
	sintheta = p0[1] / r;
	IFD printf("IntRayCircle: costheta=%g, sintheta=%g\n", costheta, sintheta);
	IFD printf("IntRayCircle: theta=%g\n", RadDeg( atan2(sintheta,costheta) ) );
	
	/* Rotate a0 and p0 by theta to place p0 on x-axis. */
	pr[0] =  costheta * p0[0] + sintheta * p0[1];
	pr[1] = -sintheta * p0[0] + costheta * p0[1];
	if ( fabs( pr[0] - r ) > FUZZ ) {
	  printf("Error in IntRayCircle: pr[0]=%g should be r=%g\n", pr[0], r);
		exit( EXIT_FAILURE );
	}
	ar[0] =  costheta * a0[0] + sintheta * a0[1];
	ar[1] = -sintheta * a0[0] + costheta * a0[1];
	IFD printf("IntRayCircle: pr = (%g,%g)\n", pr[0], pr[1] );
	IFD printf("IntRayCircle: ar = (%g,%g)\n", ar[0], ar[1] );
	
	/* Flip (reflect) point a in x-axis, to get reflected ray. */
	arf[0] =  ar[0];
	arf[1] = -ar[1];
	/* Rotate back. */
	af[0] =  costheta * arf[0] - sintheta * arf[1];
	af[1] =  sintheta * arf[0] + costheta * arf[1];
	IFD printf("IntRayCircle: arf = (%g,%g)\n", arf[0], arf[1] );
	IFD printf("IntRayCircle: af  = (%g,%g)\n", af[0], af[1] );
	/* This point af is really in the origin-centered coord. system. */
	af0[0] = af[0];
	af0[1] = af[1];
	
	/* Put points back into given coordinate system. */
	AddVec( p0,  c, p );
	AddVec( af0, c, af);
	IFD printf("IntRayCircle: p = (%g,%g), af=(%g,%g)\n", p[0], p[1], af[0], af[1] );
	return nsoln;
}
/*--------------------------------------------------------------
Assume that the circle is centered on the origin.
Returns the number of solutions, and the first point of intersection.
--------------------------------------------------------------*/
int IntRayCircle0( tPointd a, tPointd b, double r, tPointd p )
{
	double mb, disc, denom;
	double r2;
	double t1, t2, t;	/* t is the one of {t1,t2} for the 1st inter. */
	int nsoln = -1;
	int i;
	
	IFD printf("IntRayCircle0: a=(%g,%g),b=(%g,%g),r=%g\n",
		a[0],a[1],b[0],b[1],r);
	r2 = SQ(r);
	mb = SQ(a[0]) + SQ(a[1]) - a[0]*b[0] - a[1]*b[1];
	disc = -SQ(a[1]*b[0]) + 2*a[0]*a[1]*b[0]*b[1] - SQ(a[0]*b[1]) + SQ(a[0])*r2 + 
		SQ(a[1])*r2 - 2*a[0]*b[0]*r2 + SQ(b[0])*r2 - 2*a[1]*b[1]*r2 + SQ(b[1])*r2;
    
	denom = SQ(a[0]) + SQ(a[1]) - 2*a[0]*b[0] + SQ(b[0]) - 2*a[1]*b[1] + SQ(b[1]);
	IFD printf("r2=%g, mb=%g, disc=%g, denom=%g\n", r2, mb, disc, denom);
	if ( denom == 0.0 ) {
		printf("Error in IntRayCircle0: denom == 0\n");
		nsoln = 0;
		return nsoln;
	}
	if ( disc == 0.0 ) {
		IFD printf("IntRayCircle0: One solution because disc == 0\n");
		t = mb / denom;
		nsoln = 1;
	}
	if ( disc < 0.0 ) {
		IFD printf("IntRayCircle0: No solutions because negative disc\n");
		nsoln = 0;
		return nsoln;
	}
	
	t1 = ( mb - sqrt( disc ) ) / denom;
	t2 = ( mb + sqrt( disc ) ) / denom;
	IFD printf("IntRayCircle0: t1=%g, t2=%g\n", t1, t2 );
	
	if ( t1 > t2 ) 
		printf("Error in IntRayCircle0: t1 > t2 should never happen!\n");
	if ( t1 < 0.0  &&  t2 < 0.0 ) {
		IFD printf("IntRayCircle0: Both intersection on negative ray\n");
		nsoln = 0;
		return nsoln;
	}
	nsoln = 2;
	if ( t1 < 0.0 )
	  t = t2;
	else /* 0.0 < t1 < t2 */
	  t = t1;
	for( i = 0; i < DIM; i++ ) 
	  p[i] = a[i] + t * ( b[i] - a[i] );
	IFD printf("IntRayCircle0: Using t = %g: p = (%g,%g)\n", t, p[0], p[1] ); 	
	return nsoln;
}

void   MoveTo( tPointd p )
{
  printf("%10.5lf %10.5lf moveto\n", p[Xi], p[Yi] );
}

void   LineTo( tPointd p )
{
  printf("%10.5lf %10.5lf lineto\n", p[Xi], p[Yi] );
}

double   Length2( tPointd v )
{
  int   i;
  double   ss; 	
  ss = 0;

  for ( i=0; i < DIM; i++ )
    ss += v[i] * v[i];
  return ss;
}

double   RadDeg( double x )
{
  return 180 * x / M_PI;
}

/* a - b -> c */
void   SubVec( tPointd a, tPointd b, tPointd c )
{
  int   i;
  for ( i=0; i < DIM; i++ ) {
    c[i] = a[i] - b[i];
  }
}

/* a + b -> c */
void   AddVec( tPointd a, tPointd b, tPointd c )
{
  int   i;
  for ( i=0; i < DIM; i++ ) {
    c[i] = a[i] + b[i];
  }
}

void PSHead( void )
{
  printf("%%!PS\n");
  printf(".002 .002 setlinewidth\n"); 
  printf("%% ------------ Color ----------------%%\n");
  printf("/Rcolor { 1.0 0 0.45 setrgbcolor }def\n");
}



void ReScale(void)

{
  printf("%lf %lf translate\n", transX, transY ); 
  printf("%lf %lf scale\n", scaleX, scaleY ); /* 1 -> 2 inches */
  printf("newpath\n");
}


void PSTail( void )
{
	printf("\nshowpage\n\n");
	printf("%%EOF\n");
}

/* a = b */
void PointAssign( tPointd a, tPointd b )
{
  int i;
  for ( i=0; i < DIM; i++ ) {
    a[i] = b[i];
  }
}

void DrawSegments( void ) {
  for (int j = 0; j<NS; j++)
    {
      MoveTo(sStart[j]);
      LineTo(sEnd[j]);
    }
  printf("stroke\n");
  printf(".00 .00 setlinewidth\n");
}

/*Use PrintInfo() to print any useful information in the .ps file*/

void PrintInfo(void){

  printf("/Times-Roman findfont   %% Get the basic font\n");
  printf("%lf scalefont            %% Scale the font to 20 points\n", 11/scaleX);
  printf("setfont                 %% Make it the current font\n newpath                 %% Start a new path\n");
  printf("%lf %lf moveto\n", -transX/scaleX+MARGIN/(2*scaleX), -transY/scaleY+MARGIN/(2*scaleY));
  printf("(There are %d segments, %d rays; starts at point %lf, to %lf.) show\n",NS, NR, startpoint[0], endpoint[0]);  	
}

double SqDist(tPointd a, tPointd b)
{
	return((a[0]-b[0])*(a[0]-b[0])+(a[1]-b[1])*(a[1]-b[1]));
}

double Dist(tPointd a, tPointd b)
{
	return sqrt(((a[0]-b[0])*(a[0]-b[0])+(a[1]-b[1])*(a[1]-b[1])));
}



